@@ -4,135 +4,186 @@
 #' directory.
 #'
 #' @param verbose Should details of the search for Scala and Java be provided?
-#'   Or, if an rscala bridge is provided instead of a logical, the function
+#'   Or, if a Scala bridge is provided instead of a logical, the function
 #'   returns a list of details associated with the supplied bridge.
 #' @param reconfig If \code{TRUE}, the script \code{~/.rscala/config.R} is
 #'   rewritten based on a new search for Scala and Java.  If \code{FALSE}, the
 #'   previous configuration is sourced from the script
 #'   \code{~/.rscala/config.R}.  If \code{"live"}, a new search is performed,
-#'   but the results do not overwrite the previous configuration script.  Finally,
-#'   the value set here is superceded by the value of the environment variable
-#'   \code{RSCALA_RECONFIG}, if it exists.
+#'   but the results do not overwrite the previous configuration script.
+#'   \code{"offline"} is the same as \code{"live"}, except no software is ever
+#'   downloaded.
+#'   Finally, the value set here is superceded by the value of the environment
+#'   variable \code{RSCALA_RECONFIG}, if it exists.
 #' @param download A character vector which may be length-zero or whose elements
 #'   are any combination of \code{"java"}, \code{"scala"}, or \code{"sbt"}. Or,
-#'   \code{TRUE} denotes all three.  The indicated software will be installed at
+#'   \code{TRUE} denotes all three.  The indicated software will be installed in
 #'   "~/.rscala".
 #' @param require.sbt Should SBT be required, downloading and installing it in
 #'   '~/.rscala/sbt' if necessary?
 #'
 #' @return Returns a list of details of the Scala and Java binaries.
-#' @references {David B. Dahl (2018). “Integration of R and Scala Using rscala.”
-#'   Journal of Statistical Software, in editing. https://www.jstatsoft.org}
+#' @references {David B. Dahl (2019). "Integration of R and Scala Using rscala."
+#'   Journal of Statistical Software, 92:4, 1-18. https://www.jstatsoft.org}
 #' @export
 #' @examples \donttest{
-#'
 #' scalaConfig()
 #' }
+#' 
 scalaConfig <- function(verbose=TRUE, reconfig=FALSE, download=character(0), require.sbt=FALSE) {
   if ( inherits(verbose,"rscalaBridge") ) return(attr(verbose,"details")$config)
-  if ( ( length(download) > 0 ) && ( download == TRUE ) ) download <- c("java","scala","sbt")
+  if ( isTRUE(download) ) download <- c("java","scala","sbt")
   if ( length(setdiff(download,c("java","scala","sbt"))) > 0 ) stop('Invalid element in "download" argument.')
+  download <- tolower(download)
   download.java <- "java" %in% download
   download.scala <- "scala" %in% download
   download.sbt <- "sbt" %in% download
   if ( Sys.getenv("RSCALA_RECONFIG") != "" ) reconfig <- Sys.getenv("RSCALA_RECONFIG")
+  if ( reconfig %in% c("TRUE","FALSE") ) reconfig <- as.logical(reconfig)
+  else if ( ! ( reconfig %in% c("live","offline") ) ) stop("'reconfig' is misspecified.")
   consent <- identical(reconfig,TRUE) || download.java || download.scala || download.sbt
-  installPath <- path.expand(file.path("~",".rscala"))
-  dependsPath <- if ( Sys.getenv("RSCALA_BUILDING") != "" ) file.path(getwd(),"inst","dependencies") else ""
+  installPath <- path.expand(if ( Sys.getenv("RSCALA_HOME") != "" ) Sys.getenv("RSCALA_HOME") else file.path("~",".rscala"))
+  dependsPath <- if ( identical(Sys.getenv("R_INSTALL_PKG"),"rscala") ) file.path(Sys.getenv("R_PACKAGE_DIR"),"dependencies") else ""
   offerInstall <- function(msg) {
-    if ( !identical(reconfig,"live") && interactive() ) {
+    if ( ! ( reconfig %in% c("live","offline") ) && interactive() ) {
       while ( TRUE ) {
         cat(msg,"\n")
-        response <- toupper(trimws(readline(prompt="Would you like to install now? [Y/n] ")))
+        response <- toupper(trimws(readline(prompt=paste0("Do you want to install here: ",installPath,"?  [Y/n] "))))
         if ( response == "N" ) return(FALSE)
         if ( response %in% c("Y","") ) return(TRUE)
       }
     } else FALSE
   }
-  configPath  <- file.path(installPath,"config.R")
-  if ( identical(reconfig,FALSE) && file.exists(configPath) && !download.java && !download.scala && !download.sbt ) {
-    if ( verbose ) cat(paste0("\nRead existing configuration file: ",configPath,"\n\n"))
+  configPath <- file.path(installPath,"config.R")
+  if ( file.exists(configPath) && ! consent && reconfig == FALSE) {
+    if ( verbose ) cat(paste0("\nRead configuration file from rscala home: ",configPath,"\n\n"))
     source(configPath,chdir=TRUE,local=TRUE)
-    if ( is.null(config$format) || ( config$format < 4L ) || ( ! all(file.exists(c(config$javaHome,config$scalaHome,config$javaCmd,config$scalaCmd))) ) || ( is.null(config$sbtCmd) && require.sbt ) || ( ! is.null(config$sbtCmd) && ! file.exists(config$sbtCmd) ) ) {
-      if ( verbose ) cat("The 'config.R' is out-of-date.  Reconfiguring...\n")
+    if ( isConfigBroken(config, require.sbt) ) {
+      if ( verbose ) cat("The configuration is out-of-date.  Reconfiguring...\n")
       unlink(configPath)
       scalaConfig(verbose, reconfig, download, require.sbt)
     } else config
   } else {
+    tmpdir <- file.path(tempdir(),"rscala")
     if ( download.java ) installSoftware(installPath,"java",verbose=verbose)
     javaConf <- findExecutable("java","Java",installPath,javaSpecifics,verbose)
+    stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nJava is not found!  Please run 'rscala::scalaConfig(download=\"java\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
     if ( is.null(javaConf) ) {
+      if ( download.java ) stop(stopMsg)
       if ( verbose ) cat("\n")
-      consent2 <- offerInstall(paste0("Java and Scala are not found.")) 
-      stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nJava is not found!  Please run 'rscala::scalaConfig(download=\"java\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
-      if ( consent2 ) {
+      if ( ! ( reconfig %in% c("live","offline") ) && interactive() ) {
+        if ( ! ( consent || offerInstall("Java is not found.") ) ) stop(stopMsg)
+        consent <- TRUE
         installSoftware(installPath,"java",verbose=verbose)
         javaConf <- findExecutable("java","Java",installPath,javaSpecifics,verbose)
         if ( is.null(javaConf) ) stop(stopMsg)
       } else {
-        if ( dependsPath != "" ) {
-          installSoftware(dependsPath,"java",verbose=verbose)
-          javaConf <- findExecutable("java","Java",dependsPath,javaSpecifics,verbose)
-          if ( is.null(javaConf) ) stop(stopMsg)
+        if ( reconfig == "offline" ) {
+          stop(stopMsg)
+        } else {
+          if ( dependsPath != "" ) {
+            installSoftware(dependsPath,"java",verbose=verbose)
+            javaConf <- findExecutable("java","Java",installPath,javaSpecifics,verbose)
+            if ( is.null(javaConf) ) stop(stopMsg)
+          } else {
+            installSoftware(tmpdir,"java",verbose=verbose)
+            javaConf <- findExecutable("java","Java",installPath,javaSpecifics,verbose)
+            if ( is.null(javaConf) ) stop(stopMsg)
+            stopMsg <- "<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nJava was downloaded to a temporary directory.  To make permanent, please run 'rscala::scalaConfig(download=\"java\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
+            cat(stopMsg)
+          }
         }
-        else stop(stopMsg)
       }
-      consent <- consent || consent2
     }
     if ( download.scala ) installSoftware(installPath,"scala",verbose=verbose)
     scalaSpecifics2 <- function(x,y) scalaSpecifics(x,javaConf,y)
     scalaConf <- findExecutable("scala","Scala",installPath,scalaSpecifics2,verbose)
+    stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nScala is not found!  Please run 'rscala::scalaConfig(download=\"scala\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
     if ( is.null(scalaConf) ) {
+      if ( download.scala ) stop(stopMsg)
       if ( verbose ) cat("\n")
-      consent2 <- consent || offerInstall(paste0("Scala is not found.")) 
-      stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nScala is not found!  Please run 'rscala::scalaConfig(download=\"scala\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
-      if ( consent2 ) {
+      if ( ! ( reconfig %in% c("live","offline") ) && interactive() ) {
+        if ( ! ( consent || offerInstall("Scala is not found.") ) ) stop(stopMsg)
+        consent <- TRUE
         installSoftware(installPath,"scala",verbose=verbose)
         scalaConf <- findExecutable("scala","Scala",installPath,scalaSpecifics2,verbose)
         if ( is.null(scalaConf) ) stop(stopMsg)
       } else {
-        if ( dependsPath != "" ) {
-          installSoftware(dependsPath,"scala",verbose=verbose)
-          scalaConf <- findExecutable("scala","Scala",dependsPath,scalaSpecifics2,verbose)
-          if ( is.null(scalaConf) ) stop(stopMsg)
+        if ( reconfig == "offline" ) {
+          stop(stopMsg)
+        } else {
+          if ( dependsPath != "" ) {
+            installSoftware(dependsPath,"scala",verbose=verbose)
+            scalaConf <- findExecutable("scala","Scala",installPath,scalaSpecifics2,verbose)
+            if ( is.null(scalaConf) ) stop(stopMsg)
+          } else {
+            installSoftware(tmpdir,"scala",verbose=verbose)
+            scalaConf <- findExecutable("scala","Scala",installPath,scalaSpecifics2,verbose)
+            if ( is.null(scalaConf) ) stop(stopMsg)
+            stopMsg <- "<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nScala was downloaded to a temporary directory.  To make permanent, please run 'rscala::scalaConfig(download=\"scala\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
+            cat(stopMsg)
+          }
         }
-        else stop(stopMsg)
       }
-      consent <- consent || consent2
-    }
-    if ( ( javaConf$javaArchitecture == 32 ) && ( osBit() == 64 ) ) {
-      warning("32-bit Java is paired with a 64-bit operating system.  To access more memory, please run 'scalaConfig(download=\"java\")'.")
     }
     config <- c(format=4L,scalaConf,javaConf)
     if ( download.sbt ) installSoftware(installPath,"sbt",verbose=verbose)
     sbtSpecifics <- function(x,y) list(sbtCmd=x)
     sbtConf <- findExecutable("sbt","SBT",installPath,sbtSpecifics,verbose)
+    stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nSBT is not found!  Please run 'rscala::scalaConfig(download=\"sbt\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
     if ( is.null(sbtConf) && require.sbt ) {
+      if ( download.sbt ) stop(stopMsg)
       if ( verbose ) cat("\n")
-      consent2 <- consent || offerInstall(paste0("SBT is not found.")) 
-      stopMsg <- "\n\n<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nSBT is not found!  Please run 'rscala::scalaConfig(download=\"sbt\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
-      if ( consent2 ) {
+      if ( ! ( reconfig %in% c("live","offline") ) && interactive() ) {
+        if ( ! ( consent || offerInstall("SBT is not found.") ) ) stop(stopMsg)
+        consent <- TRUE
         installSoftware(installPath,"sbt",verbose=verbose)
         sbtConf <- findExecutable("sbt","SBT",installPath,sbtSpecifics,verbose)
-        if ( is.null(sbtConf) )  stop(stopMsg)
-      } else stop(stopMsg)
-      consent <- consent || consent2      
+        if ( is.null(sbtConf) ) stop(stopMsg)
+      } else {
+        if ( reconfig == "offline" ) {
+          stop(stopMsg)
+        } else {
+          if ( dependsPath != "" ) {
+            installSoftware(dependsPath,"sbt",verbose=verbose)
+            sbtConf <- findExecutable("sbt","SBT",installPath,sbtSpecifics,verbose)
+            if ( is.null(sbtConf) ) stop(stopMsg)          
+          } else {
+            installSoftware(tmpdir,"sbt",verbose=verbose)
+            sbtConf <- findExecutable("sbt","SBT",installPath,sbtSpecifics,verbose)
+            if ( is.null(sbtConf) ) stop(stopMsg)
+            stopMsg <- "<<<<<<<<<<\n<<<<<<<<<<\n<<<<<<<<<<\n\nSBT was downloaded to a temporary directory.  To make permanent, please run 'rscala::scalaConfig(download=\"sbt\")'\n\n>>>>>>>>>>\n>>>>>>>>>>\n>>>>>>>>>>\n"
+            cat(stopMsg)
+          }
+        }
+      }
     }
     config <- c(config,sbtConf)
     if ( !consent && verbose ) cat("\n")
-    writeConfig <- consent || offerInstall(paste0("File '",configPath,"' is not found or is out-of-date."))
-    if ( writeConfig ) {
-      dir.create(installPath,showWarnings=FALSE,recursive=TRUE)
-      outFile <- file(configPath,open="w")
-      dump("config",file=outFile)
-      close(outFile)
-      if ( verbose ) cat(paste0("\nWrote configuration file: ",configPath,"\n\n"))
-    } else if ( verbose ) cat("\n")
+    writeConfig <- consent || offerInstall(paste0("Configuration file is not found (or is out-of-date)."))
+    if ( writeConfig ) writeConfigFile(configPath, config, verbose)
+    else if ( verbose ) cat("\n")
     config
   }
 }
 
-findExecutable <- function(mode,prettyMode,installPath,mapper,verbose=TRUE) {  ## Mimic how the 'scala' script finds Java.
+isConfigBroken <- function(config, require.sbt) {
+  result <- try(silent=TRUE,
+    ( is.null(config$format) || ( config$format < 4L ) || ( ! all(file.exists(c(config$javaHome,config$scalaHome,config$javaCmd,config$scalaCmd))) ) || ( is.null(config$sbtCmd) && require.sbt ) || ( ! is.null(config$sbtCmd) && ! file.exists(config$sbtCmd) ) )
+  )
+  inherits(result,"try-error") || result
+}
+
+writeConfigFile <- function(configPath, config, verbose) {
+  installPath <- dirname(configPath)
+  dir.create(installPath,showWarnings=FALSE,recursive=TRUE)
+  outFile <- file(configPath,open="w")
+  dump("config",file=outFile)
+  close(outFile)
+  if ( verbose ) cat(paste0("\nWrote configuration file: ",configPath,"\n\n"))
+}
+
+findExecutable <- function(mode,prettyMode,installPath,mapper,verbose=TRUE) {
   tryCandidate <- function(candidate) {
     if ( ! is.null(candidate) && length(candidate) == 1 && candidate != "" && file.exists(candidate) ) {
       if ( verbose ) cat(paste0("  Success with ",label,".\n"))
@@ -149,7 +200,7 @@ findExecutable <- function(mode,prettyMode,installPath,mapper,verbose=TRUE) {  #
   allCaps <- toupper(mode)
   if ( verbose ) cat(paste0("\nSearching the system for ",prettyMode,".\n"))
   ###
-  label <- "directory"
+  label <- "rscala home directory"
   regex <- sprintf("%s%s$",mode,if ( .Platform$OS.type == "windows" ) "(\\.exe|\\.bat)" else "")
   candidates <- list.files(installPath,paste0("^",regex),recursive=TRUE)
   candidates <- candidates[grepl(sprintf("^%s/(.*/|)bin/%s",mode,regex),candidates)]
@@ -172,28 +223,37 @@ findExecutable <- function(mode,prettyMode,installPath,mapper,verbose=TRUE) {  #
     conf <- tryCandidate(if ( ! inherits(home,"try-error") && ( home != "" ) ) file.path(home,"bin",mode) else "")
     if ( ! is.null(conf) ) return(conf)
   } else {
+    ###
     label <- "PATH environment variable"
     conf <- tryCandidate(Sys.which(mode)[[mode]])
     if ( ! is.null(conf) ) return(conf)
+    ###
+    if ( mode == "java" ) {
+      label <- paste0("R CMD config JAVA")
+      path <- tryCatch(system2(file.path(R.home("bin"),"R"),c("CMD","config","JAVA"),stdout=TRUE,stderr=FALSE), warning=function(e) NULL, error=function(e) NULL)
+      conf <- tryCandidate(path)
+      if ( ! is.null(conf) ) return(conf)
+    }
   }
   ###
-  if ( mode == "java" ) {
-    label <- paste0("R CMD config JAVA")
-    path <- tryCatch(system2(file.path(R.home("bin"),"R"),c("CMD","config","JAVA"),stdout=TRUE,stderr=FALSE), warning=function(e) NULL, error=function(e) NULL)
-    conf <- tryCandidate(path)
-    if ( ! is.null(conf) ) return(conf)
-  } 
+  label <- "rscala package directory"
+  regex <- sprintf("%s%s$",mode,if ( .Platform$OS.type == "windows" ) "(\\.exe|\\.bat)" else "")
+  pkgDir <- if ( identical(Sys.getenv("R_INSTALL_PKG"),"rscala") ) Sys.getenv("R_PACKAGE_DIR") else system.file(package="rscala")
+  dependsPath <- file.path(pkgDir,"dependencies")
+  candidates <- list.files(dependsPath,paste0("^",regex),recursive=TRUE)
+  candidates <- candidates[grepl(sprintf("^%s/(.*/|)bin/%s",mode,regex),candidates)]
+  if ( length(candidates) > 1 ) candidates <- candidates[which.min(nchar(candidates))]
+  conf <- tryCandidate(file.path(dependsPath,candidates))
+  if ( ! is.null(conf) ) return(conf)
   ###
-  if ( Sys.getenv("RSCALA_BUILDING") == "" ) {
-    label <- "package build directory"
-    regex <- sprintf("%s%s$",mode,if ( .Platform$OS.type == "windows" ) "(\\.exe|\\.bat)" else "")
-    dependsPath <- file.path(system.file(package="rscala"),"dependencies")
-    candidates <- list.files(dependsPath,paste0("^",regex),recursive=TRUE)
-    candidates <- candidates[grepl(sprintf("^%s/(.*/|)bin/%s",mode,regex),candidates)]
-    if ( length(candidates) > 1 ) candidates <- candidates[which.min(nchar(candidates))]
-    conf <- tryCandidate(file.path(dependsPath,candidates))
-    if ( ! is.null(conf) ) return(conf)    
-  }
+  label <- "rscala temporary directory"
+  regex <- sprintf("%s%s$",mode,if ( .Platform$OS.type == "windows" ) "(\\.exe|\\.bat)" else "")
+  installPath <- file.path(tempdir(),"rscala")
+  candidates <- list.files(installPath,paste0("^",regex),recursive=TRUE)
+  candidates <- candidates[grepl(sprintf("^%s/(.*/|)bin/%s",mode,regex),candidates)]
+  if ( length(candidates) > 1 ) candidates <- candidates[which.min(nchar(candidates))]
+  conf <- tryCandidate(file.path(installPath,candidates))
+  if ( ! is.null(conf) ) return(conf)
   NULL
 }
 
@@ -207,7 +267,7 @@ extractArchive <- function(archivePath, parentDirectory, directoryName) {
   if ( file.exists(tmpInstallDirectory) ) unlink(tmpInstallDirectory, TRUE, TRUE)
   dir.create(tmpInstallDirectory, showWarnings=FALSE)
   if ( endsWith(archivePath,".zip") ) utils::unzip(archivePath, exdir=tmpInstallDirectory, unzip="internal")
-  else if ( endsWith(archivePath,".tar.gz") || endsWith(archivePath,".tgz") ) utils::untar(archivePath, exdir=tmpInstallDirectory, tar="internal")
+  else if ( endsWith(archivePath,".tar.gz") || endsWith(archivePath,".tgz") || endsWith(archivePath,".gz") ) utils::untar(archivePath, exdir=tmpInstallDirectory, tar="internal")
   else stop(paste0("Unrecognized file type for ", archivePath))
   home <- list.files(tmpInstallDirectory, full.names=TRUE)
   if ( length(home) != 1 ) stop(paste0("Expected only one directory in ",tmpInstallDirectory))
@@ -218,20 +278,23 @@ extractArchive <- function(archivePath, parentDirectory, directoryName) {
   status <- file.rename(home, finalPath)
   if ( ! status ) stop(paste0("Problem renaming ", home, " to ", finalPath))
   unlink(tmpInstallDirectory, TRUE, TRUE)
+  filesInBin <- list.files(file.path(finalPath,"bin"),full.names=TRUE)
+  Sys.chmod(filesInBin, "0755")
   finalPath
 }
 
 #' @importFrom utils download.file
+#' @importFrom tools md5sum
 #' 
-installSoftware <- function(installPath, software, version, os, bit, verbose=FALSE, downloadFailureCount=0, extractFailureCount=0) {
+installSoftware <- function(installPath, software, version, os, arch, verbose=FALSE, downloadFailureCount=0, extractFailureCount=0) {
   sel <- urls$software == software
   if ( missing(version) ) {
     version <- if ( software == "java" ) {
-      if ( Sys.getenv("RSCALA_VERIFY_JAVA_VERSION","") != "" )  Sys.getenv("RSCALA_VERIFY_JAVA_VERSION","") else "8"
+      if ( Sys.getenv("RSCALA_JAVA_VERSION","") != "" )  Sys.getenv("RSCALA_JAVA_VERSION","") else "17"
     } else if ( software == "scala" ) {
-      if ( Sys.getenv("RSCALA_VERIFY_SCALA_VERSION","") != "" ) Sys.getenv("RSCALA_VERIFY_SCALA_VERSION","") else "2.12"
+      if ( Sys.getenv("RSCALA_SCALA_VERSION","") != "" ) Sys.getenv("RSCALA_SCALA_VERSION","") else "2.13"
     } else if ( software == "sbt" ) {
-      if ( Sys.getenv("RSCALA_VERIFY_SBT_VERSION","") != "" ) Sys.getenv("RSCALA_VERIFY_SBT_VERSION","") else "1.2"
+      if ( Sys.getenv("RSCALA_SBT_VERSION","") != "" ) Sys.getenv("RSCALA_SBT_VERSION","") else "1.8"
     } else NULL
   }
   sel <- sel & (urls$version == version)
@@ -239,29 +302,34 @@ installSoftware <- function(installPath, software, version, os, bit, verbose=FAL
     os <-  if ( software == "java" ) osType() else "any"
   }
   sel <- sel & (urls$os == os)
-  if ( missing(bit) ) {
-    bit <- if ( software == "java" ) {
-      if ( Sys.getenv("RSCALA_VERIFY_JAVA_BIT","") != "" ) Sys.getenv("RSCALA_VERIFY_JAVA_BIT","") else osBit()
+  if ( missing(arch) ) {
+    arch <- if ( software == "java" ) {
+      if ( Sys.getenv("RSCALA_JAVA_ARCH","") != "" ) Sys.getenv("RSCALA_JAVA_ARCH","") else R.Version()$arch
     } else "any"
   }
-  sel <- sel & (urls$bit == bit)
+  sel <- if ( arch == "any" ) sel else {
+    sel & sapply(urls$arch, function(re) grepl(re,arch))
+  }
   urls2 <- urls[sel, ]
-  candidates <- urls2[order(as.numeric(urls2$priority),decreasing=TRUE),"url"]
+  urls3 <- urls2[order(as.numeric(urls2$priority),decreasing=TRUE), ]
   if ( verbose ) {
-    len <- length(candidates)
+    len <- nrow(urls3)
     if ( len == 1 ) cat(paste0("There is 1 candidate.\n\n"))
     else cat(paste0("There are ",len," candidates.\n\n"))
   }
-  if ( missing(downloadFailureCount) && ( Sys.getenv("RSCALA_VERIFY_DOWNLOAD_FAILURE_COUNT","") != "" ) ) downloadFailureCount <- as.integer(Sys.getenv("RSCALA_VERIFY_DOWNLOAD_FAILURE_COUNT",""))
-  if ( missing(extractFailureCount)  && ( Sys.getenv("RSCALA_VERIFY_EXTRACT_FAILURE_COUNT", "") != "" ) ) extractFailureCount  <- as.integer(Sys.getenv("RSCALA_VERIFY_EXTRACT_FAILURE_COUNT", ""))
+  if ( missing(downloadFailureCount) && ( Sys.getenv("RSCALA_DOWNLOAD_FAILURE_COUNT","") != "" ) ) downloadFailureCount <- as.integer(Sys.getenv("RSCALA_DOWNLOAD_FAILURE_COUNT",""))
+  if ( missing(extractFailureCount)  && ( Sys.getenv("RSCALA_EXTRACT_FAILURE_COUNT", "") != "" ) ) extractFailureCount  <- as.integer(Sys.getenv("RSCALA_EXTRACT_FAILURE_COUNT", ""))
   dfc <- efc <- 0
-  for ( candidate in candidates ) {
-    archivePath <- file.path(tempdir(), basename(candidate))
-    result <- try(download.file(candidate, archivePath), silent=TRUE)
+  for ( candidate in seq_len(nrow(urls3)) ) {
+    archivePath <- file.path(tempdir(), basename(urls3$url[candidate]))
+    result <- try(download.file(urls3$url[candidate], archivePath, mode="wb"), silent=TRUE)
     if ( inherits(result,"try-error") || ( result != 0 ) || ( dfc < downloadFailureCount ) ) {
       dfc <- dfc + 1
       next
     }
+    if ( tools::md5sum(archivePath) != urls3$md5sum[candidate] ) {
+      stop("Security error: md5sum do not match.")
+    }
     finalPath <- try(extractArchive(archivePath, installPath, software), silent=TRUE)
     unlink(archivePath,FALSE,TRUE)
     if ( inherits(finalPath,"try-error") || ( efc < extractFailureCount ) ) {
@@ -291,7 +359,9 @@ javaSpecifics <- function(javaCmd,verbose) {
   bit <- if ( any(grepl('.*64-?[bB]it.*$',response)) ||
               any(grepl('*amd64.*$',response)) ||
               any(grepl('.*GraalVM.*',response)) ) 64 else 32
-  list(javaCmd=javaCmd, javaMajorVersion=versionNumber, javaArchitecture=bit)
+  result <- list(javaCmd=javaCmd, javaMajorVersion=versionNumber, javaArchitecture=bit)
+  if ( verbose ) print(result)
+  result
 }
 
 setJavaEnv <- function(javaConf) {
@@ -300,7 +370,7 @@ setJavaEnv <- function(javaConf) {
   if ( oldJAVACMD  == "--unset--" ) oldJAVACMD  <- NULL
   if ( oldJAVAHOME == "--unset--" ) oldJAVAHOME <- NULL
   if ( is.null(javaConf$javaCmd)  ) Sys.unsetenv("JAVACMD")   else Sys.setenv(JAVACMD= javaConf$javaCmd)
-  if ( is.null(javaConf$javaHome) ) Sys.unsetenv("JAVA_HOME") else Sys.setenv(JAVAHOME=javaConf$javaHome)
+  if ( is.null(javaConf$javaHome) ) Sys.unsetenv("JAVA_HOME") else Sys.setenv(JAVA_HOME=javaConf$javaHome)
   list(javaCmd=oldJAVACMD,javaHome=oldJAVAHOME)
 }
 
@@ -314,13 +384,17 @@ scalaSpecifics <- function(scalaCmd,javaConf,verbose) {
   oldJavaEnv <- setJavaEnv(javaConf)
   info <- tryCatch({
     system2(scalaCmd,c("-nobootcp","-nc","-e",shQuote('import util.Properties._; println(Seq(versionNumberString,scalaHome,javaHome).mkString(lineSeparator))')),stdout=TRUE,stderr=FALSE)
-   }, warning=function(e) "")
+  }, warning=function(e) "", error=function(e) "")
   setJavaEnv(oldJavaEnv)
-  fullVersion <- info[1]
-  majorVersion <- scalaMajorVersion(fullVersion)
-  if ( majorVersion == "" ) majorVersion <- "?"
+  if ( is.null(info[1]) || is.na(info[1]) ) {
+    fullVersion <- "?"
+    majorVersion <- "?"
+  } else {
+    fullVersion <- info[1]
+    majorVersion <- scalaMajorVersion(fullVersion)
+  }
   supportedVersions <- names(scalaVersionJARs())
-  if ( ( length(supportedVersions) > 0 ) && ! ( majorVersion %in% supportedVersions ) ) paste0("unsupported Scala version: ",majorVersion)
+  result <- if ( ( length(supportedVersions) > 0 ) && ! ( majorVersion %in% supportedVersions ) ) paste0("unsupported Scala version: ",majorVersion)
   else {
     if ( ( majorVersion == "2.11" ) && ( javaConf$javaMajorVersion > 8 ) ) {
       sprintf("Scala %s is not supported on Java %s.",majorVersion,javaConf$javaMajorVersion)
@@ -328,27 +402,17 @@ scalaSpecifics <- function(scalaCmd,javaConf,verbose) {
       list(scalaHome=info[2], scalaCmd=scalaCmd, scalaMajorVersion=majorVersion, scalaFullVersion=fullVersion, javaHome=info[3])
     }
   }
+  if ( verbose ) print(result)
+  result
 }
 
 verifyDownloads <- function() {
   scalaConfig(download=c("java","scala","sbt"))
-  for ( version in c("11","8") ) {
-    for ( efc in 0:0 ) {
-      Sys.setenv(RSCALA_VERIFY_EXTRACT_FAILURE_COUNT=efc)
-      Sys.setenv(RSCALA_VERIFY_JAVA_VERSION=version)
-      Sys.setenv(RSCALA_VERIFY_JAVA_BIT=32)
-      cat(paste0("----------\nefc=",efc,", software='java', version=",version,", bit=32\n"))
-      scalaConfig(download="java")
-      s <- scala()
-      cat(s * '"OKAY"',"\n\n")
-      close(s)
-    }
-  }
-  for ( version in c("11","8") ) {
-    for ( efc in 0:2 ) {
-      Sys.setenv(RSCALA_VERIFY_EXTRACT_FAILURE_COUNT=efc)
-      Sys.setenv(RSCALA_VERIFY_JAVA_VERSION=version)
-      Sys.setenv(RSCALA_VERIFY_JAVA_BIT=64)
+  for ( version in c("17","11","8") ) {
+    for ( efc in 0:1 ) {
+      Sys.setenv(RSCALA_EXTRACT_FAILURE_COUNT=efc)
+      Sys.setenv(RSCALA_JAVA_VERSION=version)
+      Sys.setenv(RSCALA_JAVA_ARCH="x86_64")
       cat(paste0("----------\nefc=",efc,", software='java', version=",version,"\n"))
       scalaConfig(download="java")
       s <- scala()
@@ -356,10 +420,10 @@ verifyDownloads <- function() {
       close(s)
     }
   }
-  for ( version in c("2.11","2.13.0-RC1","2.12") ) {
+  for ( version in c("2.11","2.12","2.13") ) {
     for ( efc in 0:1 ) {
-      Sys.setenv(RSCALA_VERIFY_EXTRACT_FAILURE_COUNT=efc)
-      Sys.setenv(RSCALA_VERIFY_SCALA_VERSION=version)
+      Sys.setenv(RSCALA_EXTRACT_FAILURE_COUNT=efc)
+      Sys.setenv(RSCALA_SCALA_VERSION=version)
       cat(paste0("----------\nefc=",efc,", software='scala', version=",version,"\n"))
       scalaConfig(download="scala")
       s <- scala()
@@ -367,15 +431,15 @@ verifyDownloads <- function() {
       close(s)
     }
   }
-  for ( version in c("1.2") ) {
+  for ( version in c("1.8") ) {
     for ( efc in 0:1 ) {
-      Sys.setenv(RSCALA_VERIFY_EXTRACT_FAILURE_COUNT=efc)
-      Sys.setenv(RSCALA_VERIFY_SBT_VERSION=version)
+      Sys.setenv(RSCALA_EXTRACT_FAILURE_COUNT=efc)
+      Sys.setenv(RSCALA_SBT_VERSION=version)
       cat(paste0("----------\nefc=",efc,", software='sbt', version=",version,"\n"))
       scalaConfig(download="sbt")
       scalaSBT("sbtVersion")
       cat("\n")
     }
   }
-  Sys.unsetenv(c("RSCALA_VERIFY_EXTRACT_FAILURE_COUNT","RSCALA_VERIFY_JAVA_VERSION","RSCALA_VERIFY_SCALA_VERSION","RSCALA_VERIFY_SBT_VERSION","RSCALA_VERIFY_JAVA_BIT"))
+  Sys.unsetenv(c("RSCALA_EXTRACT_FAILURE_COUNT","RSCALA_JAVA_VERSION","RSCALA_SCALA_VERSION","RSCALA_SBT_VERSION","RSCALA_JAVA_ARCH"))
 }
