@@ -1,3 +1,3 @@
 close(s)
-close(s2)
+# close(s2)
 cat("Stopping instance. **************************\n")
